#include "dfosfor.h"
#include <math.h>
#include <intrin.h>	//__cpuid()
#include <smmintrin.h>	// Fehlerbereinigte Version ohne Semikolon nach #define
//SSE4.1 wird für _mm_floor_ps() benötigt.
#pragma intrinsic(memcpy,memset,memcmp)

// Globale Variablen:
HINSTANCE ghInstance;
HANDLE ghConOut;
TCHAR szTitle[64];
bool avx_avail;
LARGE_INTEGER pf;

static size_t alignup(size_t x,size_t ali) {return x+ali-1 & ~(ali-1);}

template<class T,unsigned ALIGN>class _declspec(align(64)) Array2D{
 Array2D(const Array2D&o);	// =delete
 void operator=(const Array2D&o); // =delete
 static int alignup(int w) {return (int)::alignup(w*sizeof(T),ALIGN)/sizeof(T);}
protected:
 int width,height;
 T*a;
 T garbage[2];
public:
 size_t size() const {return size_t(width)*size_t(height);}
 Array2D(int w,int h):width(w=alignup(w)),height(h) {
  size_t l=w*h*sizeof(T);
  void*v=_aligned_malloc(l,ALIGN);
  if (v) memset(v,0,l);
  a=reinterpret_cast<T*>(v);
 }
 ~Array2D() {_aligned_free(a);}
 T&getPixel(int x,int y) {
  return (unsigned)x<(unsigned)width && (unsigned)y<(unsigned)height ? a[y*width+x] : garbage[0]=0;
 }
// wie getPixel, aber bei Bereichsüberprüfung 1 freies Pixel rechts garantieren
 __forceinline T*getPixel1(int x,int y) {
  return (unsigned)x<unsigned(width-1) && (unsigned)y<(unsigned)height ? a+y*width+x : (T*)memset(garbage,0,sizeof garbage);
 }
 T&operator() (int x,int y) {return getPixel(x,y);}
 T&addToPixel(int x,int y,T a) {return getPixel(x,y)+=a;}
 T*data() {return a;}
 const T*data() const {return a;}
 BITMAPINFO*getBitmapInfo(BITMAPINFOHEADER&bih) const{
  memset(&bih,0,sizeof bih);
  bih.biSize=sizeof bih;
  bih.biWidth=width;
  bih.biHeight=-height;
  bih.biPlanes=1;
  bih.biBitCount=sizeof(T)<<3;
  return (BITMAPINFO*)&bih;
 }
};

template<class T>T limit(T v, T mini, T maxi) {return v<mini?mini:v>maxi?maxi:v;}

template<class T>void swap(T&v1,T&v2) {T t=v1; v1=v2; v2=t;}

template<class T>T f_abs(T v) {return v<0?-v:v;}

__forceinline int lrint(float f){
 return _mm_cvt_ss2si(_mm_set_ss(f));
}

// Horizontales Maximum im Ergebnis (Lo-Float)
__forceinline __m128 hmax(__m128 f) {
      f=_mm_max_ps(f,_mm_movehl_ps(f,f));
 return _mm_max_ss(f,_mm_shuffle_ps(f,f,_MM_SHUFFLE(0,0,0,1)));
}

__forceinline float hmaxf(__m128 f) {
 f=hmax(f);
 float r;
 _mm_store_ss(&r,f);
 return r;
}

extern "C"
#if _MSC_VER < 1200
 _declspec(naked)
#endif
void _cdecl _fltused() {};	// Linker ruhig stellen

static float displaygamma=1;	// sinnvoll 0..2

__forceinline __m128 gamma(__m128 v) {
 __m128 h=_mm_set_ps1(0.5f);	// Hälfte
 __m128 w=_mm_sub_ps(v,h);	// Mittenzentrierter Wertebereich ±0,5
 __m128 q=_mm_mul_ps(w,w);	// Quadrat, also 0..0,25
 __m128 b=_mm_sub_ps(_mm_set_ps1(0.25f),q);	// Buckel = Erhöhung in der Mitte
 return _mm_add_ps(v,_mm_mul_ps(_mm_set_ps1(displaygamma),b));
}

// Teufel nochmal, der Flaschenhals ist nun addPixel1()
void addColors(Array2D<RGBQUAD,64>&dst,const Array2D<float,64>&src,COLORREF c,float f) {
//assert(dst.width==src.width && dst.height==src.height)
 size_t rcx=dst.size()>>2;
 if (!rcx) return;
 __m128i*dp=reinterpret_cast<__m128i*>(dst.data());
 const __m128*sp=reinterpret_cast<const __m128*>(src.data());
 __m128i ci={char(c>>16),char(c>>8),char(c)};
 __m128 cf=_mm_cvtepi32_ps(_mm_cvtepu8_epi32(ci));
 __m128 f4=_mm_set_ps1(f);
 do{
  __m128 g4=gamma(_mm_mul_ps(*sp++,f4));
  __m128 g0=_mm_shuffle_ps(g4,g4,_MM_SHUFFLE(0,0,0,0));	// spread index 0
  __m128 g1=_mm_shuffle_ps(g4,g4,_MM_SHUFFLE(1,1,1,1));
  g0=_mm_mul_ps(g0,cf);
  g1=_mm_mul_ps(g1,cf);
  __m128i h0=_mm_cvtps_epi32(g0);
  __m128i h1=_mm_cvtps_epi32(g1);
  __m128i il=_mm_packs_epi32(h0,h1);
  g0=_mm_shuffle_ps(g4,g4,_MM_SHUFFLE(2,2,2,2));
  g1=_mm_shuffle_ps(g4,g4,_MM_SHUFFLE(3,3,3,3));
  g0=_mm_mul_ps(g0,cf);
  g1=_mm_mul_ps(g1,cf);
  h0=_mm_cvtps_epi32(g0);
  h1=_mm_cvtps_epi32(g1);
  __m128i ih=_mm_packs_epi32(h0,h1);
  __m128i ic=_mm_packus_epi16(il,ih);
  *dp=_mm_adds_epu8(*dp,ic);
  dp++;
 }while(--rcx);
}

class _declspec(align(64)) Fosfor:public Array2D<float,64>{
protected:
 __m128 last;	// Lo-Float32[2] = letzte Koordinate
 __m128 ones;
 mutable bool xyswap;
 void line(__m128 q);
 bool clip(__m128&q) const;
 void set(__m128 p, __m128 a) {
  if (!a.m128_f32[0]) return;
// In der Programmlogik ist x Führungsgröße und ganzzahlig.
// y ist gebrochen.
  __m128 b=_mm_floor_ps(p),
	 f=_mm_sub_ps(p,b),
	 g=_mm_sub_ps(ones,f);
  __m128i bi=_mm_cvtps_epi32(b);
  if (xyswap) bi=_mm_shuffle_epi32(bi,_MM_SHUFFLE(0,0,0,1));
  setMax(addToPixel(bi.m128i_i32[0],bi.m128i_i32[1],_mm_mul_ss(a,_mm_shuffle_ps(f,f,_MM_SHUFFLE(0,0,0,1))).m128_f32[0]));
  if (!g.m128_f32[1]) return;
  setMax(addToPixel(bi.m128i_i32[0],bi.m128i_i32[1]+1,_mm_mul_ss(a,_mm_shuffle_ps(g,g,_MM_SHUFFLE(0,0,0,1))).m128_f32[0]));
 }
 int getW() const {return xyswap?height:width;}
 int getH() const {return xyswap?width:height;}
public:
 union{
  __m128 max4;
#if _MSC_VER > 1900
  __m256 max8;
#endif
 };
// Fügt 4 Pixeln im Quadrat die vier Werte von <f> hinzu.
 __forceinline void addFourPixel(__m128i p, __m128 f) {
  __declspec(align(16)) int xy[4];
  _mm_storel_epi64((__m128i*)xy,p);
  __m64 *a=(__m64*)getPixel1(xy[0],xy[1]),	// Not aligned by design!
	*b=(__m64*)getPixel1(xy[0],xy[1]+1);
  __m128 v;
  v=_mm_loadl_pi(f,a);	// gather
  v=_mm_loadh_pi(v,b);
  v=_mm_add_ps(v,f);
  _mm_storel_pi(a,v);	// scatter
  _mm_storeh_pi(b,v);
  max4=_mm_max_ps(max4,v);
 }

#if _MSC_VER > 1900
 // Als AVX-Version für 2 Quadrate: Lohnt nicht so recht wegen Overlap-Problem!
 __forceinline void addFourPixel_2(__m128i p, __m256 f) {
  __declspec(align(16)) int xy[4];
  _mm_storel_epi64((__m128i*)xy,p);
  __int64 *a=(__int64*)getPixel1(xy[0],xy[1]),	// Not aligned by design!
	  *b=(__int64*)getPixel1(xy[0],xy[1]+1),
	  *c=(__int64*)getPixel1(xy[2],xy[3]),	// may overlap with a and b! This will give erraneous result!!
	  *d=(__int64*)getPixel1(xy[2],xy[3]+1);
  __m256 v=_mm256_castsi256_ps(_mm256_set_epi64x(*d,*c,*b,*a));	// gather
  v=_mm256_add_ps(v,f);
  __declspec(align(32)) __int64 vi[4];
  _mm256_store_si256((__m256i*)vi,_mm256_castps_si256(v));	// scatter
  *a=vi[0]; *b=vi[1]; *c=vi[2]; *d=vi[3];
  max8=_mm256_max_ps(max8,v);
 }
#endif

 void setMax(float v) {max4=_mm_max_ss(max4,_mm_set_ss(v));}
 Fosfor(int w,int h):Array2D<float,64>(w,h) {
  memset(&max4,0,sizeof max4);
  ones=_mm_set_ps1(1);
 }
 void moveTo(__m128 p) {last=p;}
 void lineTo(__m128 p) {
  __m128 q=_mm_shuffle_ps(p,last,_MM_SHUFFLE(3,2,1,0));
  if (clip(q)) line(q);
  moveTo(p);
 }
 inline void addPoint0(__m128 p);	// draw (unaliased) point of width 0
 inline void addPoint1(__m128 p);	// draw (aliased) point of width 1
 inline void addPoint1_2(__m128 p2) {addPoint1(p2); addPoint1(_mm_shuffle_ps(p2,p2,_MM_SHUFFLE(1,0,3,2)));}
 __m128 maximum() const;
#if _MSC_VER > 1900
 inline void addPoint1_4(__m256 p4);	// Mit avx
#endif
};

bool Fosfor::clip(__m128&q) const{
 static const DWORD mask=0x7FFFFFFF;
 __m128 p=_mm_shuffle_ps(q,q,_MM_SHUFFLE(1,0,3,2)),
	d=_mm_sub_ps(p,q),
	da=_mm_and_ps(_mm_load_ps1(reinterpret_cast<const float*>(&mask)),d);
 if (!da.m128_i64) return false;
 if (xyswap=da.m128_f32[1]>da.m128_f32[0]) {
  q=_mm_shuffle_ps(q,q,_MM_SHUFFLE(2,3,0,1));	// swap X and Y
  p=_mm_shuffle_ps(q,q,_MM_SHUFFLE(1,0,3,2));	// recalculate
  d=_mm_sub_ps(p,q);				// recalculate
 }
 if (d.m128_f32[0]<0) {
  q=_mm_shuffle_ps(q,q,_MM_SHUFFLE(1,0,3,2));	// swap source and target
  p=_mm_shuffle_ps(q,q,_MM_SHUFFLE(1,0,3,2));	// recalculate
  d=_mm_sub_ps(p,q);				// recalculate
 }
 float a=d.m128_f32[1]/d.m128_f32[0];
 if (q.m128_f32[0]>=getW()) return false;
 if (q.m128_f32[2]<0) return false;
 if (q.m128_f32[0]<0) {q.m128_f32[1]-=q.m128_f32[0]*a; q.m128_f32[0]=0;}
/*TODO
 if (p[1].x>=getW()) {
  float r=getW()-1E-4f;
  p[1].y-=(p[1].x-r)*a; p[1].x=r;
 }
 if (a>=0) {
  if (p[0].y>=getH()) return false;
  if (p[1].y<0) return false;
  if (p[0].y<0) {p[0].x-=p[0].y/a; p[0].y=0;}
  if (p[1].y>=getH()) {
   float b=getH()-1E-4f;
   p[1].x-=(p[1].y-b)/a; p[1].y=b;
  }
 }else if (a<0) {
  if (p[0].y<0) return false;
  if (p[1].y>=getH()) return false;
  if (p[0].y>=getH()) {
   float b=getH()-1E-4f;
   p[0].x-=(p[0].y-b)/a; p[0].y=b;
  }
  if (p[1].y<0) {p[1].x-=p[1].y/a; p[1].y=0;}
 }
*/
 return q.m128_f32[2]>=q.m128_f32[0];
}

// Hier p[1].x>p[0].x (Rechtslauf) und |Anstieg|<1 (flach)
// Kann sein dass beide x-Werte im gleichen Pixel liegen!
void Fosfor::line(__m128 q) {
 __m128 p=_mm_shuffle_ps(q,q,_MM_SHUFFLE(1,0,3,2)),
	d=_mm_sub_ps(p,q),
	a=_mm_div_ss(_mm_shuffle_ps(d,d,_MM_SHUFFLE(0,0,0,1)),d);	// Anstieg
/*TODO
 float ix0=ceilf(p[0].x),
       ix1=floorf(p[1].x),
       dx=p[1].x-p[0].x,
       dy=p[1].y-p[0].y,
       a=dy/dx;
 if (ix0>ix1) set(lrint(ix1),(p[0].y+p[1].y)*0.5f,1.0f);	// Beide x im gleichen Pixel
 else{
  float dx0=ix0-p[0].x,
	dx1=p[1].x-ix1,
	fy0=p[0].y+dx0*a,	// Geraden-Durchstoßpunkt an Pixelgrenze
        fy1=p[1].y-dx1*a;
  int intx0=lrint(ix0),
      intx1=lrint(ix1);
  set(intx0-1,(p[0].y+fy0)*0.5f,dx0/dx);
  for (int i=0; i<intx1-intx0; i++) {
   set(intx0+i,fy0+(i+0.5f)*a,1.0f/dx);
  }
  set(intx1,(fy1+p[1].y)*0.5f,dx1/dx);
 }
*/
}

void Fosfor::addPoint0(__m128 p) {
 int xy[4];
 _mm_storel_epi64((__m128i*)xy,_mm_cvtps_epi32(_mm_floor_ps(p)));
 setMax(addToPixel(xy[0],xy[1],1));
}

void Fosfor::addPoint1(__m128 p) {
 __m128	b=_mm_floor_ps(p),
	f=_mm_sub_ps(p,b),				// -  - fy fx	fractional part for right/bottom pixel
	g=_mm_sub_ps(ones,f),				// -  - gy gx	inverse-factional part for target pixel
	h=_mm_movelh_ps(g,f);				//fy fx gy gx
 f=_mm_mul_ps(			// fractional parts for target - right - bottom - bottom-right pixel
 	_mm_shuffle_ps(h,h,_MM_SHUFFLE(2,0,2,0)),	//fx gx fx gx
	_mm_shuffle_ps(h,h,_MM_SHUFFLE(3,3,1,1)));	//fy fy gy gy
 addFourPixel(_mm_cvtps_epi32(b),f);
}

#if _MSC_VER > 1900
void Fosfor::addPoint1_4(__m256 p4) {			// p3y p3x p2y p2x p1y p1x p0y p0x
#if 1
 addPoint1_2(_mm256_extractf128_ps(p4,0));
 addPoint1_2(_mm256_extractf128_ps(p4,1));
#else
 __m256 b=_mm256_floor_ps(p),				// b3y b3x b2y b2x b1y b1x b0y b0x
	f=_mm256_sub_ps(p,b),				// f3y f3x f2y f2x f1y f1x f0y f0x
	g=_mm256_sub_ps(_mm256_set1_ps(1),f);		// g3y g3x g2y g2x g1y g1x g0y g0x
 __m128 h,k;
 h=_mm_movelh_ps(
	_mm256_extractf128_ps(g,0),
	_mm256_extractf128_ps(f,0));
 k=_mm_mul_ps(
	_mm_shuffle_ps(h,h,_MM_SHUFFLE(2,0,2,0)),
	_mm_shuffle_ps(h,h,_MM_SHUFFLE(3,1,3,1)));
 addFourPixel(_mm_cvtps_epi32(_mm256_extractf128_ps(b,0)),k);

 h=_mm_movelh_ps(
	_mm256_extractf128_ps(g,0),
	_mm256_extractf128_ps(f,0));
 k=_mm_mul_ps(
	_mm_shuffle_ps(h,h,_MM_SHUFFLE(2,0,2,0)),
	_mm_shuffle_ps(h,h,_MM_SHUFFLE(3,1,3,1)));
 addFourPixel(_mm_cvtps_epi32(_mm256_extractf128_ps(b,0)),k);
 
 h=_mm_movelh_ps(
	_mm256_extractf128_ps(g,1),
	_mm256_extractf128_ps(f,1));
 k=_mm_mul_ps(
	_mm_shuffle_ps(h,h,_MM_SHUFFLE(2,0,2,0)),
	_mm_shuffle_ps(h,h,_MM_SHUFFLE(3,1,3,1)));
 addFourPixel(_mm_cvtps_epi32(_mm256_extractf128_ps(b,1)),k);
 
 h=_mm_movelh_ps(
	_mm256_extractf128_ps(g,1),
	_mm256_extractf128_ps(f,1));
 k=_mm_mul_ps(
	_mm_shuffle_ps(h,h,_MM_SHUFFLE(2,0,2,0)),
	_mm_shuffle_ps(h,h,_MM_SHUFFLE(3,1,3,1)));
 addFourPixel(_mm_cvtps_epi32(_mm256_extractf128_ps(b,1)),k);
#endif
}
#endif

__m128 Fosfor::maximum() const{
 __m128 max4=_mm_setzero_ps();
 const __m128*p = reinterpret_cast<const __m128*>(data());
 size_t rcx=size()>>2;	// wegen zeilenweisem Align==16 stets durch 4 teilbar!
 do max4=_mm_max_ps(max4,*p++); while(--rcx);
 return hmax(max4);
}

static COLORREF custcolors[16];
static CHOOSECOLOR cc={
 sizeof cc,
 0,
 0,
 RGB(255,255,0),
 custcolors,
 CC_ANYCOLOR|CC_FULLOPEN|CC_RGBINIT
};
/*
RGBQUAD fromColorref(COLORREF cr) {
 RGBQUAD c={BYTE(cr>>16),BYTE(cr>>8),BYTE(cr)};
 return c;
}
*/
static size_t yvallen;
static float*xyvalues;

static void calcpoints(int w) {
 size_t n=w*1000;	// Punktezahl
 if (yvallen<n) {
  size_t bytes=alignup(n*2*sizeof(float),64);
  n=bytes/2/sizeof(float);
  LARGE_INTEGER t[2];
  QueryPerformanceCounter(t);
  if (xyvalues) _aligned_free(xyvalues);
  xyvalues=(float*)_aligned_malloc(bytes,64);
  yvallen=n;
  float hf=10;		// HF-Trägerfrequenz
  float ha=80;		// HF-Amplitude (Scheitelwert)
  float nf=0.1f;	// NF-Frequenz
  float nm=0.25f;	// Modulationsgrad
  for (size_t i=0;i<n; i++) {
   float x=i*0.001f;
   xyvalues[i*2]=x;
   xyvalues[i*2+1]=ha*sinf(x*hf)*(1+nm*sinf(x*nf));
  }
  QueryPerformanceCounter(t+1);
  DebugOut("calcpoints(%d) dauerte %.2f ms, belegt %u KByte",w,double(t[1].QuadPart-t[0].QuadPart)*1E3/double(pf.QuadPart),bytes>>10);
 }
}
struct LinScale{
 float m,n;	// y = m×x+n;
 float scale(float x) const {return m*x+n;}
 float unscale(float y) const {return (y-n)/m;}
 void setm(float z, float y) {float x=unscale(y); m=z; n-=scale(x)-y;}
 void zoom(float z, float y) {setm(m*(1+z),y);}
};
static LinScale xyscale[2];

static void onPaint(HDC dc,int w,int h) {
 if (!xyvalues) return;
 size_t n=w*1000;	// Punktezahl
 LARGE_INTEGER t[4];
 Array2D<RGBQUAD,64>picture(w+1,h);
 Fosfor akku(w+1,h);
 float ym=(float)h*0.5f;
 QueryPerformanceCounter(t);
#if _MSC_VER > 1900
 if (avx_avail) {
  __m256 p0=_mm256_set_ps(ym,0,ym,0,ym,0,ym,0);
  for (size_t i=0; i<n; i+=4) {
   __m256 p4=_mm256_add_ps(_mm256_load_ps(xyvalues+i*2),p0);	// Vier Punkte, acht floats
   akku.addPoint1_4(p4);
  }
 }else
#endif
 {
  __m128 pn=_mm_set_ps(0,0,xyscale[1].n,xyscale[0].n),
	 pm=_mm_set_ps(0,0,xyscale[1].m,xyscale[0].m);
  pn=_mm_shuffle_ps(pn,pn,_MM_SHUFFLE(1,0,1,0));
  pm=_mm_shuffle_ps(pm,pm,_MM_SHUFFLE(1,0,1,0));
  for (size_t i=0; i<n; i+=2) {
   akku.addPoint1_2(_mm_add_ps(_mm_mul_ps(_mm_load_ps(xyvalues+i*2),pm),pn));	// Zwei Punkte, vier floats
  }
 }
 QueryPerformanceCounter(t+1);
 addColors(picture,akku,cc.rgbResult,1/hmaxf(akku.max4));	// 2 = Übersättigung zulassen
 QueryPerformanceCounter(t+2);
 BITMAPINFOHEADER bih;
 SetDIBitsToDevice(dc,0,0,w,h,0,0,0,h,picture.data(),picture.getBitmapInfo(bih),0);
 QueryPerformanceCounter(t+3);
 DebugOut("addPoints dauerte %.2f ms",double(t[1].QuadPart-t[0].QuadPart)*1E3/double(pf.QuadPart));
 DebugOut("addColors dauerte %.2f ms",double(t[2].QuadPart-t[1].QuadPart)*1E3/double(pf.QuadPart));
 DebugOut("SetDIBits dauerte %.2f ms",double(t[3].QuadPart-t[2].QuadPart)*1E3/double(pf.QuadPart));
}

static void InfoOut(const char*name,int infobit0) {
 DebugOut("%-7s: %s",name,infobit0&1?"Ja":"Nein");
}

static void InfoSse() {
 int info[4];
 __cpuid(info,0);
 int nIds=info[0];
 if (nIds>=1) {
  __cpuid(info,1);
  InfoOut("SSE3",info[2]);
  InfoOut("SSSE3",info[2]>>9);
  InfoOut("SSE4.1",info[2]>>19);
  InfoOut("SSE4.2",info[2]>>20);
  InfoOut("AES",info[2]>>25);
  InfoOut("AVX",info[2]>>28);
  if (info[2] & 1<<28) avx_avail=true;
 }
 if (nIds>=7) {
  __cpuid(info,7);
  InfoOut("AVX2",info[1]>>5);
 }
}

static POINTS mousepos;

static LRESULT CALLBACK WndProc(HWND Wnd, UINT msg, WPARAM wParam, LPARAM lParam) {
 switch (msg) {
  case WM_CREATE: {
   xyscale[1].m = xyscale[0].m = 1;
   QueryPerformanceFrequency(&pf);
   InfoSse();
   cc.hwndOwner=Wnd;
  }break;

  case WM_RBUTTONDOWN: {
   if (ChooseColor(&cc)) InvalidateRect(Wnd,0,FALSE);
  }break;
  
  case WM_MOUSEWHEEL: {
   POINT pt={GET_X_LPARAM(lParam),GET_Y_LPARAM(lParam)};
   ScreenToClient(Wnd,&pt);
   float z = short(HIWORD(wParam))*0.001f;
   if (!z) break;
   if (!(wParam&MK_CONTROL)) xyscale[0].zoom(z,float(pt.x));
   if (wParam&MK_SHIFT) xyscale[1].zoom(z,float(pt.y));
   InvalidateRect(Wnd,0,FALSE);
  }break;
  
  case WM_LBUTTONDOWN: {
   mousepos=*(POINTS*)&lParam;
   SetCursor(LoadCursor(0,MAKEINTRESOURCE(32654)));
   SetCapture(Wnd);
  }break;
  case WM_MOUSEMOVE: if (GetCapture()==Wnd) {
   POINTS newpos=*(POINTS*)&lParam;
   xyscale[0].n+=newpos.x-mousepos.x;
   xyscale[1].n+=newpos.y-mousepos.y;
   mousepos=newpos;
   InvalidateRect(Wnd,0,FALSE);
  }break;
  case WM_LBUTTONUP: ReleaseCapture(); break;

  case WM_COMMAND: switch (wParam) {
   case 2: DestroyWindow(Wnd); break;
  }break;
  
  case WM_SIZE: /*seltsam, das geht nicht if (wParam)*/ {
   calcpoints(GET_X_LPARAM(lParam));
   xyscale[1].n=GET_Y_LPARAM(lParam)*0.5f;
  }break;

  case WM_PAINT:{
   PAINTSTRUCT ps;
   BeginPaint(Wnd,&ps);
   RECT rc;
   GetClientRect(Wnd,&rc);
   onPaint(ps.hdc,rc.right,rc.bottom);
   EndPaint(Wnd,&ps);
  }break;

  case WM_DESTROY: PostQuitMessage(0); break;
  
  case WM_NCDESTROY: if (xyvalues) _aligned_free(xyvalues); break;
  
  default: return DefWindowProc(Wnd,msg,wParam,lParam);
 }
 return 0;
}

ATOM registerWnd(){
 WNDCLASS wc={
  CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS,
  WndProc,
  0,
  0,
  ghInstance,
  LoadIcon(ghInstance,MAKEINTRESOURCE(1)),
  LoadCursor(0,IDC_ARROW),
  GetStockBrush(BLACK_BRUSH),
  0,
  TEXT("dfosfor")};
 return RegisterClass(&wc);
}

void WinMainCRTStartup() {
 ghInstance=GetModuleHandle(0);
 if (AttachConsole(ATTACH_PARENT_PROCESS)) {
  ghConOut=CreateFile("CONOUT$",GENERIC_WRITE,FILE_SHARE_WRITE,0,OPEN_EXISTING,0,0);
  DebugOut("\r\033[K");	// Kommandoprompt löschen (aber wenn's geht OHNE Zeilenschaltung danach)
 }
 LoadString(0,0,szTitle,elemof(szTitle));
 registerWnd();
 ShowWindow(CreateWindow(TEXT("dfosfor"),szTitle,WS_OVERLAPPEDWINDOW,
   CW_USEDEFAULT,0,CW_USEDEFAULT,0,0,0,ghInstance,0),SW_SHOWDEFAULT);

 MSG msg;
 while (GetMessage(&msg,0,0,0))	{
  TranslateMessage(&msg);
  DispatchMessage(&msg);
 }
 ExitProcess((UINT)msg.wParam);
}
