#pragma once
// Interne Klassen fr PlotXY-Fensterklasse, C++
#include "plotxy.h"
#include "scale.h"
#include "axis.h"
#include "trace.h"

#include "gridlayout.h"
#include "../util/compat.h"

class Axis;

struct PLOTATTR{
  DWORD shrink_ms;	// Millisekunden fr Auto-Shrink der Skalen
  Color cBack, cArea, cLegend;	// alle Plotbereiche gleiche Hintergrundfarbe
// High-Byte = Transparenz. 0xFFFFFFFF = voll transparent zu GetSysColor(COLOR_WINDOW)
//			    0xFF0000xx = GetSysColor(xx), Transparenz ignorieren
  enum{
    bColorSchemeDark,
    bcbkgndScaleSysColorIndex=8,
    bcbkgndPlotSysColorIndex,
  };
  DWORD flags;
};

class PlotXY:public GridLayout,public PLOTATTR {
 friend class WindowTextParser;
 mutable HDC dc;	// kann von WM_PAINT, WM_PRINTCLIENT oder GetDC() kommen
 mutable int n_getdc;	// Zhler fr getdc(), Verschachtelungen sind erlaubt, PlotXY cacht.
public:
 HWND hPlot,hMini,hTip;
 Brush bBack,bArea;

// Traces breiten sich im GridLayout nur dort aus, wo beide Indizes ungerade sind!
 typedef vector<uniq_ptr<Trace> > traces_t;
 traces_t traces;
// Achsen breiten sich im GridLayout in allen anderen Zellen aus
 typedef vector<uniq_ptr<Axis> > axes_t;
 axes_t axes;		// alle Achsen

 PlotXY(HWND w):dc(0),n_getdc(0),
	hPlot(w),hMini(0),hTip(0),
	dirty(0),
	hCursor(LoadCursor(0,IDC_NO)) {
  span_in_plots_inc(0);
  span_in_plots_inc(1);
 }
 Axis&insertAxis(AXISATTR&);	// entnimmt side&1, entnimmt+korrigiert iplotarea, iaxis, liefert XAxis* oder YAxis*
 Axis&eraseAxis(AXISATTR&);	// entnimmt iplotarea, side&1, iaxis, liefert XAxis* oder YAxis*
 Trace*insertTrace(TRACEATTR&);
 Trace*eraseTrace(TRACEATTR&);

 Axis&assignAxis(AXISATTR&);	// findet oder erstellt eine passende Achse, ggf. unter Erstellung einer Plotarea
 Axis&assignAxis(Trace&,DWORD side);	// dito und fgt den Trace der Achse hinzu
 DWORD axisAttrOfColRow(unsigned xysel,char cr);
 int textWidth(const TCHAR*,int=-1) const;
 Metric textMetric() const;
 void arrange();
 void invalidate();			// alles (erforderlich wenn sich die Aufteilung verndert)
 void invalidate(const gridrange_t&);	// Ein oder mehrere (zusammenhngende) Grid-Bereiche
 void invalidate(char ix, char iy);	// ix, iy in Grid-Koordinaten
 void invalidate(const TRACEATTR&);	// Plotbereich des Traces (wenn sich Linienzug ndert)
 void invalidate(const AXISATTR&);	// Achsenbereich (wenn sich Skale bei konstantem Platzbedarf ndert)
 HDC getdc() const;			// entweder der DC von WM_PAINT oder vom schnden GetDC
 void releasedc() const;		// rckgngig

 enum{
  GAP=3,
  MOUSECAPTURE=20,	// Pixel Abstand zur Linie bzw. deren Endpunkt
 };
 static struct SystemSettings{
  TCHAR sDecimal[2];	// TEXT(",") oder TEXT(".")
  TCHAR sTime[4];	// TEXT(":")
  bool iMeasure;	// false = metrisch oder true = imperial
  void readFromSystem();
  bool operator==(const SystemSettings&o) const {return !memcmp(this,&o,sizeof*this);}
 }systemsettings;
 enum{
  bCheckAutorange,
  bCheckAxes,
  bArrangeAll,
 };
 LONG dirty;		// LONG damit _bittestandset usw. funktioniert
 mutable struct Mousetrack{
    char icurve,	// fr Mehrfachlinien (Oszi-Betrieb)
         itrace,	// Nur EDITABLE-Traces
	 iplotx,
	 iploty;
    int segment;	// Integer-Anteil: Polyliniensegment
    Float where;	// Gebrochener Anteil: Wo auf dem Segment, 0..1 = dazwischen, sonst ber die Enden hinaus
    Float distance;	// Vorgefundener kleinster Abstand
    bool tracking;	// wenn nahe genug ODER Mauscapture
    Mousetrack():tracking(false) {}
  }mousetrack;
  Axis::ticks_t::const_iterator findEditableLabel(gridpos_t, Point<>&) const;

  char span_in_plots(unsigned xysel) const {return gridsize(xysel)>>1;}

  void span_in_plots_inc(unsigned xysel,bool fixed=false);
// return column or row index of plot area
  static char icr(char iplotxy) {return (iplotxy<<1)+1;}
  static char icr(const AXISATTR&a) {return icr(a.iplotxy);}
// return width of column or height of row
  int wh_cr(unsigned side, char cr) const {if (cr<0) cr+=gridsize(side&1); return extents_xy[side&1].extents[cr]->get();}
  int wh_cr(const Axis&ax) const {return wh_cr(ax.side^1,irc(ax));}
// return width or height of plot area (needed for scale arrange)
  int wh(unsigned side, char iplotxy) const {return wh_cr(side,icr(iplotxy));}
  int wh(const Axis&ax, char neben=0) const {return wh_cr(ax.side,icr(ax)+neben);}
// return row or column index for a scale (always on an outer gap
// side.0 selects either 0 = X scale or 1 = Y scale
// side.1 selects either 0 = left/top or 1 = right/bottom
// side.2 selects either 0 = growing to right/bottom, 1 = growing to left/top
// Therefore, the most typical values for the lower 3 bit are 2 for X axes and 5 for Y axes.
  char irc(unsigned side) const {return side&2?gridsize(side)-1:0;}
  char irc(const AXISATTR&a) const {return irc(a.side^1);}
// Die Rechtecke und Rnder der Plotbereiche mssen in ihrer jeweiligen Richtung gleich positioniert sein.
// Die Rnder zwischen den Plotbereichen fallen zusammen (collapse) und bilden Griffe zur Proportionsnderung.
  gridpos_t findArea(Point<>&pt) const;	// verndert <pt>
  const TCHAR*gridmovecursor(gridpos_t) const;
//  bool isDividerVariable(int);
  static Float distance(const Point<>line[2],const Point<>&pt,Float&where);	// Abstand quer zur Linie oder zu den Endpunkten
  traces_t::const_iterator findTrace(const Point<>&pt) const;	// Polylinie finden, die am nchsten kommt
  void onSetText(const TCHAR*);	// Text aus Ressource oder SetDlgItemText()/SetWindowText() verarbeiten
  bool onSize(int w, int h, bool fromOutside);

private:
  HCURSOR hCursor;

// aus Funkuhr/Karte.c
  BYTE bFromMouse:1;
  char HoverIndex;

// Client-Pixel-Verschiebung
  int shift(bool nBar) const {return 0/*(&ScrollRect.left)[nBar]-ScrollInfo[nBar].nPos*/;}
  int GenShift(bool xysel, char kode, int dx);
  void SetScale(char kode);
  static bool moveinto(int l, int r, int L, int R, int&v);
  int MoveInto(int k, bool nBar) const;
  int MoveRectInto(const RECT *rc, const POINT *pt, bool nBar) const;
// entweder wenn sich die Mausposition ndert oder sich die Karte unter der Maus verschiebt
  void MousePosChanged() {}	// setzt ggf. Statuszeile
// Fensterinhalt um <dx> und <dy> verschieben, liefert <true> wenn tatschlich verschoben wurde
  bool Scroll(int dx, int dy);
  bool HandleHScroll(char kode) {return Scroll(GenShift(SB_HORZ,kode,0),0);}
  bool HandleVScroll(char kode) {return Scroll(0,GenShift(SB_VERT,kode,0));}
// Schieben mit Tastatur oder Maus
// Liefert <true> wenn tatschlich geschoben wurde
  bool HandleShift(int dx, int dy) {return Scroll(GenShift(SB_HORZ,-1,-dx),GenShift(SB_VERT,-1,-dy));}
  void FocusTrace(int k);
// Die Tab-Taste setzt die Hover-Funktion der Maus auer Kraft
  void HandleTab(TCHAR suche);
  void SetHoverIndex(char value) {}

 void paint(HDC dc, Rect<>* =0);
 static LRESULT CALLBACK wndproc(HWND,UINT,WPARAM,LPARAM);
 LRESULT wndproc(UINT,WPARAM,LPARAM);
public:
  static BOOL init();
};
