#include "gridlayout.h"
#define _USE_MATH_DEFINES
#include <cmath>	// fabs()
#include "../util/compat.h"	// lrint()
#include <shlwapi.h>	// wnsprintf()

//ExtentList::ExtentList(int def) {
// insert(0,new ExtentFix(def));
//}

char ExtentList::makeValidSpan(char valid_i, char span) const {
 char e = span>=0 ? valid_i+span : static_cast<char>(extents.size())+span;
 if (e<valid_i) e=valid_i;
 if (e>static_cast<char>(extents.size())) e=(char)extents.size();
 return e-valid_i;
}

bool ExtentList::insert(char index,Extent*info) {
 return extents.insert(index>=0 ? extents.begin()+index : extents.end(),info);
}

RANGE<int>ExtentList::extent(char a, char l)const {
 RANGE<int> r /*= *__$ReturnUdt*/;
 a = makeValidIndex(a);
 l = makeValidSpan(a,l);
 r.e=r.a=0;
 for (Extents::const_iterator it=extents.begin(); it!=extents.end(); it++) {
  const Extent&ex=**it;
  if (a>0) {
   r.a+=ex.get();	// Breiten links von <a> zusammenzhlen
   --a;
  }else{
   if (l>0) {
    r.e+=ex.get();	// Breiten zwischen <a> und <a>+<l>-1 zusammenzhlen
    --l;
   }
   if (l<=0) break;	// Nicht weiterzhlen
  }
 }
 r.e+=r.a;		// In Bereich von Anfang bis Ende-1 umwandeln
 return r;
}

ExtentList::Extents::const_iterator ExtentList::locatePos(int&such) const {
 if (such>=0) for (Extents::const_iterator it=extents.begin(); it!=extents.end(); it++) {
  int w=(*it)->get();
  if (such<w) return it;
  such-=w;
 }
 return 0;
}

int ExtentList::getSum(cb_t cb) const{
 int r=0;
 for (Extents::const_iterator it=extents.begin(); it!=extents.end(); it++) r+=(it->get()->*cb)();
 return r;
}

double ExtentList::Cache::Entry::delta_to_keep_limit(double newsize) const{
 double correct=min-newsize;	// Ist newsize zu klein muss mit correct>0 korrigiert werden
 if (correct>0) return correct;
 correct=max-newsize;	// Ist newsize zu gro muss mit correct<0 korrigiert werden
 if (correct<0) return correct;
 return 0;
}

// Liest die Spaltenbreiten, wenn sich irgendetwas nennenswert gendert hat
bool ExtentList::Cache::load(const Extents&src,double&wg) {
 bool changed=entries.size()!=src.size();
 if (changed) entries.resize(src.size());
 Entries::iterator di=entries.begin();
 Extents::const_iterator si=src.begin();
 wg=0;
 for (;di!=entries.end(); di++, si++) {
  const Extent&ex=*si->get();
  Entry&e=*di;
  double v=ex.get();
  if (!changed && fabs(e.x-v)>1) changed=true;
  if (changed) e.x=v;
  v=ex.getMin();
  if (!changed && e.min!=v) changed=true;
  if (changed) e.min=v;
  v=ex.getMax();
  if (!changed && e.max!=v) changed=true;
  if (changed) e.max=v;
  wg+=e.weight=ex.getWeight();	// geht nicht in changed() ein!
 }
 if (!wg) for (di=entries.begin();di!=entries.end(); di++) wg+=di->weight=1.0;
 return changed;
}

// Schreibt die Spaltenbreiten als Integerwerte zurck und behlt die Gleitkommawerte
void ExtentList::Cache::commit(Extents&dst) {
 ASSERT(entries.size()==dst.size());
 Entries::const_iterator si=entries.begin();
 Extents::iterator di=dst.begin();
 for (;si!=entries.end(); si++, di++) di->get()->set(lrint(si->x));
}

// Gesamtausdehnung festlegen
bool ExtentList::set(int sz) {
 double delta=sz-get();
 if (!delta) return false;
 double wg;
 cache.load(extents,wg);
 Cache::Entries::iterator it;
// Muss wiederholt werden bis alles stimmt!!
 for (it=cache.entries.begin(); it!=cache.entries.end(); it++) {
  double weighted_delta=delta*it->weight/wg;
  if (weighted_delta) {
   double newsize = it->x+weighted_delta;
// Eine Grenbegrenzung msste das Rest-Weight ndern!
   double correct=it->delta_to_keep_limit(newsize);
   it->x=newsize+correct;
   delta-=weighted_delta+correct;
  }
  wg-=it->weight;
// Wird das verbleibende Gewicht Null muss nun auch delta Null sein!
// Das geht schief sobald Limits erreicht werden!
  ASSERT(wg || !delta);
 }
 cache.commit(extents);
// Am Ende der Grenaufteilung sollte das verbleibende delta Null sein!
 ASSERT(!delta);
// Und niemand sollte bei den Einzelobjekten dazwischenfunken
#ifdef _DEBUG
 delta=sz-get();
 if (delta) {
  DEBUG(("%s(%d): Verbleibendes Delta von %.2f bei sz=%d",__FILE__,__LINE__,delta,sz));
  debuglog();
 }
#endif
 return true;
}

void ExtentList::setMin(const int*sz) {
 for (Extents::iterator it=extents.begin(); it!=extents.end(); it++,sz++) {
  Extent&ex=**it;
  ex.setMin(*sz);
 }
}

void ExtentList::debuglog() const{
 char buf[256],*sp=buf,*se=buf+elemof(buf);
 sp+=wnsprintfA(sp,int(se-sp),"Extents = (");
 for (Extents::const_iterator it=extents.begin(); it!=extents.end(); it++) {
  sp+=wnsprintfA(sp,int(se-sp),"%d%c",it->get()->get(), it+1==extents.end() ? ')':',');
 }
 DEBUG(("%s",buf));
}

Rect<>GridLayout::extent(gridrange_t gr) const{
 Rect<> r /*= *__$ReturnUdt*/;
 RANGE<int> rng;
 rng = extents_xy[0].extent(gr.x,gr.colspan);
 r.left  = rng.a;
 r.right = rng.e;
 rng = extents_xy[1].extent(gr.y,gr.rowspan);
 r.top   = rng.a;
 r.bottom= rng.e;
 return r;
}
