#pragma once
// Interne Klassen fr PlotXY-Fensterklasse, C++
#include "../util/gdix.h"	// Float
#include "../util/compat.h"	// lrint

/****************
 * Skalierungen *
 ****************/

// Virtuelle Basisklasse
class Scale{
public:
  virtual void make(Float,Float,int,int)=0;	// Zweipunktkalibrierung
  virtual int operator()(Float) const=0;
  virtual Float unscale(int) const=0;
//virtual ~Scale() {}
};

// Nicht skalieren
class NoScale:public Scale{
  virtual void make(Float,Float,int,int) {}
  virtual int operator()(Float x) const {return (int)x;}
  virtual Float unscale(int X) const {return (Float)X;}
};

// Linear skalieren
class LinScale:public Scale{
  Float f;
  int O;
  virtual void make(Float,Float,int,int);
  virtual int operator()(Float x) const {return lrint(x*f)+O;}
  virtual Float unscale(int X) const {return Float(X-O)/f;}
public:
  LinScale(Float s=1,int o=0):f(s),O(o) {}
};

// Logarithmisch skalieren
struct LogScale:public Scale{
  Float f;
  int O;
  bool neg;
  virtual void make(Float,Float,int,int);
  virtual int operator()(Float) const;
  virtual Float unscale(int) const;
public:
  LogScale(Float s=1,int o=0,bool n=false):f(s),O(o),neg(n) {}
};
