#include "trace.h"
#include "cplotxy.h"

/*********
 * Trace *
 *********/

Trace::Trace(PlotXY&p, TRACEATTR&td):
 pl(p),
 TRACEATTR(td),		// Hier erfolgt das Raubkopieren der Datenquellen und das Kopieren des Namens
 tr_fillto(0),
#pragma warning(disable:4355)	// Verwendung von <this> in Initialisierungsliste; hier erforderlich um Referenzen zu initialisieren
 xaxis(p.assignAxis(*this,Axis::BOTTOM/*|1<<Axis::bTime*/)),
 yaxis(p.assignAxis(*this,Axis::LEFT)) {
#pragma warning(default:4355)
  flags|=VISIBLE|CHECK_XRANGE|CHECK_YRANGE;
  _bittestandset(&p.dirty,PlotXY::bCheckAxes);
}

Trace::~Trace() {pl.dirty|=1<<PlotXY::bCheckAxes;}

void Trace::calcpolys() {
 if (!(flags&1<<bNeedCalcPolys)) return;
// TODO: Der wahrscheinlichste Fall, dass die Linienschar aus 1 Linie besteht,
// sollte irgendwie gesondert behandelt werden.
#ifdef _DEBUG
 static Point<>linesave[1024];
 if (itrace==0 && lines.size()==1) {
  vector<Point<> >&lf = *lines.front();
  if (lf.size()==1024) memcpy(linesave,lf.data(),sizeof linesave);
 }
#endif
// Auch der Fall, dass die Zahl konstant ist.
 lines.clear();		// das rumt den Heap auf
 if (!xydata[0]) return;	// Keine Datenquelle
 if (!xydata[1]) return;
 int n=xydata[0]->size();
 if (n>xydata[1]->size()) n=xydata[1]->size();	// Minimum beider
 mark.resize(marker?n:0);
 uniq_ptr<points_t> curLine;
 for (int i = 0; i < n; i++) {	// Sample-Index
  Point<>pt;
  char fx=xaxis.scaleInf(xydata[0]->get(i),pt.x);
  char fy=yaxis.scaleInf(xydata[1]->get(i),pt.y);
  if (fx || fy) {
     // curLine->clip();
   lines.push_back(curLine);	// Fernpunkt: Linie abschlieen; der Raubkopierkonstruktor macht curLine.ptr zu Null
  }else{
   if (!curLine) {
    curLine = new points_t;	// Linie beginnen
    curLine->reserve(n-i);	// Der wahrscheinlichste Fall, dass alle (restlichen) Punkte keine Fernpunkte sind
   }
   curLine->push_back(pt);
  }
  if (marker) {
   mark[i] = pt;
  }
 }
 if (curLine) {
    // curLine->clip();
  lines.push_back(curLine);	// abschlieen
 }
#ifdef _DEBUG
 if (itrace==0) {
  ASSERT(memcmp(linesave,lines.front()->data(),sizeof linesave));	// soll sich ndern!
//  InvalidateRect(pl.hPlot,0,TRUE);
 }
#endif
 flags&=~(1<<bNeedCalcPolys);
}

Range<Float>&Trace::getRange(size_t xysel) {
 Range<Float>&r=cachedRanges[xysel];
 if (flags&1<<(bNeedUpdateRange+xysel)
 && xydata[xysel]
 && xydata[xysel]->getRange(r))
	flags&=~(1<<(bNeedUpdateRange+xysel));
 return r;
}

char Trace::measure(Margin<>&padding) const{
 HDC dc=pl.getdc();
 int lw = linewidth; if (!lw) lw=1;	// Nullbreite = 1 Pixel
 int lw2= lw>>1;		// dieser Teil ist links und oben vom gedachten Punkt
 padding.expand(lw2,lw2,lw-lw2,lw-lw2);
 if (marker) {
  SIZE sz,sz2;	// Auch bei der Textvermessung wird von der Mitte ausgegangen. Die Hhe ist bertrieben.
  GetTextExtentPoint32(dc,&marker,1,&sz);
  sz2.cx=sz.cx>>1;
  sz2.cy=sz.cy>>1;
  padding.expand(sz2.cx,sz2.cy,sz.cx-sz2.cx,sz.cy-sz2.cy);
 }
 pl.releasedc();
 return 0;
}

/* Ein Trace besteht aus:
 - Polygon der Minima und Maxima
 - Polygon oder Polyline der Mittelwerte
 */
void Trace::paintPoly(HDC dc) {
 if (!visible()) return;
 if (!fill.size()) return;
 if (!bfill) bfill=Brush::create(fcolor);	// TODO: Farbe mit Pseudo-Transparenz
 SelectPen(dc,GetStockPen(NULL_PEN));
 SelectBrush(dc,bfill);
 Polygon(dc,(POINT*)fill.data(),(int)fill.size());
}
void Trace::paintLine(HDC dc) {
 if (!visible()) return;
 if (!lines.size()) return;
 if (!pline) pline=Pen::create(linestyle,linewidth,tcolor);
 SelectPen(dc,pline);
 for (lines_t::iterator il=lines.begin(); il!=lines.end(); il++) {
  points_t&line=**il;
  Polyline(dc,(POINT*)line.data(),(int)line.size());
 }
}
void Trace::paintMark(HDC dc) {
 if (!visible()) return;
 if (!mark.size()) return;
 SetTextColor(dc,mcolor);
 if (!fmark) fmark=Font::create(fdmark);
 SelectFont(dc,fmark);
 Metric tm(dc);
 int dy=(tm.ascent-tm.intlead)>>1;	// Halbe Buchstabenhhe runter
 SetTextAlign(dc,TA_CENTER|TA_BASELINE);
 for (points_t::iterator ppt=mark.begin(); ppt!=mark.end(); ppt++) {
  TextOut(dc,ppt->x,ppt->y+dy,&marker,1);
 }
}

