// Objektorientierte GDI-Kapselung
#pragma once

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>		// GetEffectiveClientRect
#include <tchar.h>		// _tcstoul()
#include "../util/cstr.h"	// cstr<> for class Font::Data

#ifndef Float
#  define Float float		// Wertekoordinaten: hier 4-Byte-Gleitkomma
#endif
//typedef LONG Int;		// Anzeigekoordinaten, muss zum Elementtyp von Rect (= RECT) passen

#if _MSC_VER < 1400
# define override		// "override" is not a keyword in MSVC6
#endif

#if _MSC_VER >= 1900
#  include <limits>
#  define FINF (std::numeric_limits<Float>::infinity())
#else
const unsigned FNAN=0xFFFFFFFF;	// Bei double anpassen!	unsigned __int64 0xFFFFFFFFFFFFFFFF
const unsigned FINF=0x7F800000;	//			unsigned __int64 0x7FF0000000000000
#define FNAN (*(Float*)&FNAN)	// #defines rekursieren nicht
#define FINF (*(Float*)&FINF)
#endif

// Stringformat-Konstanten fr ANSI<=>Unicode-Konvertierung (ohne UTF-8 faktisch nur fr Debugoutput brauchbar)
#ifdef UNICODE
# define SWW L"s"	// %s specifier for WCHAR format and WCHAR argument
# define SWT L"s"	// %s specifier for WCHAR format and TCHAR argument
# define SWS L"S"	// %s specifier for WCHAR format and  CHAR argument
# define STW TEXT("s")	// %s specifier for TCHAR format and WCHAR argument
# define STT TEXT("s")	// %s specifier for TCHAR format and TCHAR argument
# define STS TEXT("S")	// %s specifier for TCHAR format and  CHAR argument
# define SSW "S"	// %s specifier for  CHAR format and WCHAR argument
# define SST "S"	// %s specifier for  CHAR format and TCHAR argument
# define SSS "s"	// %s specifier for  CHAR format and  CHAR argument (classic)
#else
# define SWW L"s"	// %s specifier for WCHAR format and WCHAR argument
# define SWT L"S"	// %s specifier for WCHAR format and TCHAR argument
# define SWS L"S"	// %s specifier for WCHAR format and  CHAR argument
# define STW TEXT("S")	// %s specifier for TCHAR format and WCHAR argument
# define STT TEXT("s")	// %s specifier for TCHAR format and TCHAR argument
# define STS TEXT("s")	// %s specifier for TCHAR format and  CHAR argument
# define SSW "S"	// %s specifier for  CHAR format and WCHAR argument
# define SST "s"	// %s specifier for  CHAR format and TCHAR argument
# define SSS "s"	// %s specifier for  CHAR format and  CHAR argument (classic)
#endif

const int INAN=0x80000000;
const int IINF=0x7FFFFFFF;
/* Fast schon langweilige Objektifizierungen = Kapselklassen von GDI-Krimskrams
   Kann so leichter auf GDIplus umgestellt werden
 */

template<class T>struct RANGE{	// Als einfache Datenstruktur, um kein __$ReturnUdt zu generieren
 T a,e;
};
// RANGE hat ein grundstzliches Problem mit Integerzahlen:
// Soll e inklusive oder exklusive sein?
// In IntRange wird es exklusive verwendet.

// T = Integer oder Pointer. Durchlauf exklusive Ende.
template<class T,class DIFF_T=T>struct IntRange:public RANGE<T>{
 DIFF_T size() const {return DIFF_T(e-a);}
 void assign(T na,T nb) {a=na,b=nb;}
 bool inside(T v) const {return a<=v && v<e;}
};

#pragma warning(disable:4201)	// nameless struct/union
template<class T=int> struct Point{	// knnte auch von complex<T> erben
 union{
  struct{
   T x,y;
  };
  T xy[2];	// fr Arrayzugriff
 };
// Spezialisierung fr int, short und char sollte den Vergleich ber den nchstgreren Datentyp machen, aber wie?
 bool operator==(const Point<T>&o) const {return x==o.x && y==o.y;}
 bool operator!=(const Point<T>&o) const {return !operator==(o);}
// Point() {}
// Point(T xx,T yy):x(xx),y(yy) {}
// Nur zulssig fr sizeof(T)==4
 operator POINT*() {return reinterpret_cast<POINT*>(this);}
 operator const POINT*() const {return reinterpret_cast<const POINT*>(this);}
};

// 240228: Erbt NICHT von RECT weil jenes LONG (= lstig) und nicht int verwendet!
// Mit dem Template-Parameter sind auch handliche <short>- und <char>-Rechtecke mglich
// Sowie <float>, aber die sind nicht mit denen in gdiplus vergleichbar, weil mit Breite und Hhe definiert
// Oder wren die sowieso als Vorlage besser??
// Warum zur Hlle der Indexoperator nicht funktioniert, wenn man Spezialisierungen verwendet!
template<class T=int> struct Rect{	// Bei GdiPlus: RectF? Nee! RectF definiert sich gegenber RECT durch Breite und Hhe!
 union{
  struct{
   T left,top,right,bottom;
  };
  T ltrb[4];
 };
 Rect() {}	// Nicht-initialisierender Standardkonstruktor
 operator SIZE() const {SIZE sz={width(),height()}; return sz;}
 void operator=(const RECT&r) {
  left  =(T)r.left;
  top   =(T)r.top;
  right =(T)r.right;
  bottom=(T)r.bottom;
 }
 void setRECT(RECT&r) const{
  r.left  =left;
  r.top   =top;
  r.right =right;
  r.bottom=bottom;
 }
 T width()  const {return right-left;}
 T height() const {return bottom-top;}
 T center() const {return (left+right)>>1;}	// Problem: >>-Operator fr float nicht definiert!
 T middle() const {return (top+bottom)>>1;}	// Und /2 hat bei Integer andere (ungnstige) Auswirkung bei negativen Zahlen.
 T byIndex(size_t i) const {
  if (i<4) return (&left)[i];
  switch (i) {
   case 4: return width();
   case 5: return height();
   case 6: return -width();
   case 7: return -height();
   case 8: return center();
   case 9: return middle();
   default: return INAN;
  }
 }
//  operator T*() {return &left;}	// geht nicht bei Spezialisierung, verdammt!
//  operator const T*() const {return &left;}
//  T operator[](size_t i) const {return byIndex(i);}
 T&at(size_t i) {
  if (i<4) return (&left)[i];
  return *(T*)0;
 }
 T addTo(size_t i,T v) {
  if (i>=4) return 0;
  return (&left)[i]+=v;
 }
 void prepareExpand() {
  top=left=IINF;
  bottom=right=-IINF;
 }
 void zero() {memset(this,0,sizeof*this);}
 char operator+=(const Rect&o) {	// UnionRect, liefert als Ergebnis an welcher Seite etwas erweitert wurde
  return char(toMin(left,o.left))	// Kein extra Test auf leeres Rechteck!
	|char(toMin(top,o.top))<<1
	|char(toMax(right,o.right))<<2
	|char(toMax(bottom,o.bottom))<<3;}
// Die folgenden Funktionen sind nur korrekt fr sizeof(T)==4
 operator RECT*() {return (RECT*)this;}
 operator const RECT*() const {return (const RECT*)this;}
 Rect(const RECT&R) {::CopyRect(*this,&R);}
 Rect(HWND w) {::GetWindowRect(w,*this);}
 Rect(HWND w, bool) {::GetClientRect(w,*this);}
 Rect(HWND w, const int*info) {GetEffectiveClientRect(w,*this,const_cast<int*>(info));}
 void inflate(int dx,int dy) {::InflateRect(*this,dx,dy);}	// aufblasen
 void deflate(int dx,int dy) {inflate(-dx,-dy);}		// verkleinern
 void offset(int dx,int dy) {::OffsetRect(*this,dx,dy);}	// versetzen
 bool inside(const Point<>&pt) const {return !!::PtInRect(*this,*pt);}
 bool intersect(Rect<T>&rd,const Rect<T>&r2) const {return !!IntersectRect(rd,*this,r2);}
 bool intersect(const Rect&r2) const {Rect rd; return intersect(r2);}
};
/*
// Spezialisierung auf 32 Bit erlaubt Cast zwischen RECT und Rect<>
template<>struct Rect<int>{
  int left,top,right,bottom;
//  const Rect&operator=(const RECT&o) {::CopyRect(*this,&o); return this;}
//  operator int*() {return &left;}
//  const int&operator[](size_t i) const {return (&left)[i];}
  int width() const {return right-left;}	// Ich verstehe nicht warum es dieser Zeilen braucht!
  int height() const {return bottom-top;}
};
*/

template<class T=int>struct Margin{
  T l,t,r,b;
  operator T*() {return &l;}
  operator const T*() const {return &l;}
  Margin(T ll=0,T tt=0,T rr=0,T bb=0):l(ll),t(tt),r(rr),b(bb) {}
  void init() {memset(this,0,sizeof*this);}
  void init(T ll,T tt,T rr,T bb) {l=ll;t=tt;r=rr;b=bb;}
  void expand(const Margin&m) {expand(m.l,m.t,m.r,m.b);}
  void expand(T ll,T tt,T rr,T bb) {
	toMax(l,ll);
	toMax(t,tt);
	toMax(r,rr);
	toMax(b,bb);}
  T getW() const {return l+r;}	// Gesamtrand horizontal
  T getH() const {return t+b;}	// Gesamtrand vertikal
  T at(size_t i) const {switch (i) {
	case 0:case 1:case 2:case 3:return (&l)[i];
	case 4: return getW();
	case 5: return getH();
	default: return INAN;}}
  T addTo(size_t i,T v) {
    if (i>=4) return 0;
    return (&l)[i]+=v;
  }
};

#if _MSC_VER >= 1400
#  define NODEFAULT =delete	// bei neuerem C++
#else
#  define NODEFAULT
#endif

struct Color{
 enum{
  cTransparent = COLORREF(DWORD(-1))
 };
 union{
  COLORREF c;
  BYTE rgba[4];
  struct{
   BYTE r,g,b,a;	// Windows-kompatibel Rot im LoByte, a = 0xFF = transparent (nicht wie sonst opak)
  };			// Ich frage mich stndig, wieso man das bei Microsoft nicht gleich so gemacht hat!
 };
// Mache <*this> opak unter Anwendung der Hintergrundfarbe <o>
 void opak(Color o) {	// Anwendung auf opaker Hintergrundfarbe o, Ergebnis = opak
  r = mix(r,o.r,a);
  g = mix(g,o.g,a);
  b = mix(b,o.b,a);
  a = 0;
 }
// Mische Komponente <c> mit Komponente <o> im Verhltnis 255-<a> : <a>
// D.h. bei a == 0 kommt <c> und bei a == 255 kommt <o> heraus.
// Keine der Komponenten ist vormultipliziert (premultiplied)!
 static BYTE mix(BYTE c, BYTE o, BYTE a) {return (c*BYTE(255-a)+o*a+128U)/255U;}
 static BYTE nibble2byte(BYTE n) {return n<<4|n&15;}
// Konstruktoren
 Color(COLORREF cr=cTransparent):c(cr) {}		// Transparent initialisieren
 Color(BYTE r, BYTE g, BYTE b, BYTE aa=0):c(RGB(r,g,b)),a(aa) {}
 Color(int syscolorindex,bool resolve):c(resolve ? GetSysColor(syscolorindex): 0xFF000000|syscolorindex) {}
 Color operator=(COLORREF cr) {return c=cr;}
 operator COLORREF() const {return applyOn().c;}	// Das Ganze zu COLORREF casten
 operator COLORREF() {return applyOn().c;}		// ohne const die Farbe abspeichern
 operator bool() const {return c!=cTransparent;}	// Transparent ist wie nicht definiert
 bool operator!() const {return !operator bool();}	// Nicht den COLORREF-Cast benutzen
 Color applyOn(Color b = cTransparent) const{
  Color r(c);	// Kopie zum Zurckgeben
  if (HIWORD(r.c)==0xFF00) r.c = GetSysColor(LOWORD(r.c));
  else if (r.a) {	// aktuelle Farbe hat Transparenzkomponente?
   r.opak(b.applyOn(0xFF000000|COLOR_WINDOW)); // Eine(!) Rekursion, dann in-place modifizieren
  }
  return r;
 }
 Color applyOn(Color b = cTransparent) {return *this=const_cast<const Color*>(this)->applyOn(b);}

 static Color fromString(const TCHAR*s) {	// windowtextparser-kompatibel bis zum ersten Zeichen < 0x20
  Color c;
  switch (*s) {
   case '$': c = _tcstoul(++s,0,0) | 0xFF000000; break;	// Index dezimal fr GetSysColor
   case '#': {
    TCHAR*e;
    Color t = _tcstoul(++s,&e,16);
    switch (e-s) {		// konvertierte Hexadezimalziffern
     case 3:	c.r = nibble2byte(BYTE(t.c>>8));	// Nibble-Angabe r-g-b
		c.g = nibble2byte(BYTE(t.c>>4));
		c.b = nibble2byte(BYTE(t.c));
		c.a = 0;
     break;
#if 0
     case 4:	c.r = nibble2byte(BYTE(t.c>>12));	// Nibble-Angabe r-g-b-a
		c.g = nibble2byte(BYTE(t.c>>8));
		c.b = nibble2byte(BYTE(t.c>>4));
		c.a = nibble2byte(BYTE(t.c));
     break;
#endif
     case 6:	c.r = BYTE(t.c>>16);		// HTML-Angabe r-g-b
		c.g = BYTE(t.c>>8);
		c.b = BYTE(t.c);
		c.a = 0;
     break;
     default:	c = t;			// Alle anderen Lngen: a-b-g-r
    }
   }break;
   default: c = fromName(s);
  }
  return c;
 }
 static Color fromName(const TCHAR*s, TCHAR**e=0) {
  return _tcstoul(s,e,0);	// erst mal nur als Zahl, (englische?, deutsche? aus Ressource?) Farbnamen spter
 }
};
#pragma warning(default:4201)

class GdiObj{
protected:
 HGDIOBJ h;
public:
 GdiObj(HGDIOBJ o=0):h(o) {}		// GDI-Handle in Besitz nehmen
 GdiObj(const GdiObj&o):h(o.h) {const_cast<GdiObj&>(o).release();}	// Raubkopie
 void operator=(const GdiObj&o) {swap(h,const_cast<GdiObj&>(o).h);} // Ressourcentausch statt Zuweisung
 void operator=(HGDIOBJ o) {if (h) DeleteObject(h); h=o;}
 HGDIOBJ release() {HGDIOBJ r=h; h=0; return r;} // avoid deletion before leaving scope, in case it's handled otherwise
 ~GdiObj() {if (h) DeleteObject(h);}
 operator HGDIOBJ() const {return h;}	// wenn HGDIOBJ gefragt wird
 operator bool() const {return !!h;}
 bool operator!() const {return !h;}
};

struct Pen:public GdiObj{
 Pen(HPEN pen=0):GdiObj(pen) {}	// Achtung! Destruktor zerstrt den Stift!
 Pen(Pen&pen):GdiObj(pen) {pen.h=0;}
// Pen(const Pen&)NODEFAULT;	// Dreierregel: Kopieren und Zuweisen verboten
// Pen&operator=(const Pen&)NODEFAULT;
 static HPEN create(int width, Color color) {return CreatePen(PS_SOLID,width,color);}
 static HPEN create(int style,int width,Color color) {return CreatePen(style,width,color);}
 Pen(int width,Color color):GdiObj(create(width,color)) {}
 Pen(int style,int width,Color color):GdiObj(create(style,width,color)) {}
 void operator=(HPEN p) {if (h) DeletePen(h); h=p;}
 operator HPEN() const {return (HPEN)h;}	// wenn HPEN gefragt wird
};

struct Brush:public GdiObj{
 Brush(HBRUSH br=0):GdiObj(br) {}	// Achtung! Destruktor zerstrt den Pinsel!
 static HBRUSH create(Color color) {return CreateSolidBrush(color);}
 static HBRUSH create(int hatch, Color color) {CreateHatchBrush(hatch,color);}
 Brush(Color color):GdiObj(create(color)) {}
 Brush(int hatch, Color color):GdiObj(create(hatch,color)) {}
 void operator=(HBRUSH br) {if (h) DeleteBrush(h); h=br;}
 operator HBRUSH() const {return (HBRUSH)h;}	// wenn HBRUSH gefragt wird
};

struct Font:public GdiObj{
  enum{	// Font-Flags
    bold=4,
    italic=8,
    underline=16,
    strikeout=32,
    symbol=64,
    cleartype=128,
  };
  struct Data{
    cstr<>name;	// will auto-initialize with empty string
    int size;
    DWORD flags;
    int rotation;
    Data() {}
    Data(const TCHAR*n,int s,int f,int r):name(n),size(s),flags(f),rotation(r) {}
    HFONT operator()() const {
      return CreateFont(size,0,rotation,rotation,
       flags&bold ? FW_BOLD : FW_NORMAL,
       flags&italic ? TRUE : FALSE,
       flags&underline ? TRUE : FALSE,
       flags&strikeout ? TRUE : FALSE,
       flags&symbol ? SYMBOL_CHARSET : ANSI_CHARSET,
       OUT_TT_PRECIS,
       CLIP_DEFAULT_PRECIS,
#ifdef CLEARTYPE_QUALITY
       flags&cleartype ? CLEARTYPE_QUALITY : 
#endif
       PROOF_QUALITY,
       DEFAULT_PITCH|FF_DONTCARE,
       name.c_str());
    }
    Data&operator()(HFONT f) {
      LOGFONT lf;
      GetObject(f,sizeof lf,&lf);
      rotation=lf.lfOrientation;
      flags=0;
      if (lf.lfWeight>FW_NORMAL) flags|=bold;
      if (lf.lfItalic) flags|=italic;
      if (lf.lfUnderline) flags|=underline;
      if (lf.lfStrikeOut) flags|=strikeout;
      if (lf.lfCharSet==SYMBOL_CHARSET) flags|=symbol;
#ifdef CLEARTYPE_QUALITY
      if (lf.lfOutPrecision==CLEARTYPE_QUALITY) flags|=cleartype;
#endif
      size=lf.lfHeight;
      name=lf.lfFaceName;	// malloc + copy
    }
  };
  Font(HFONT f=0):GdiObj(f) {}	// Achtung! Destruktor zerstrt die Schrift!
  Font(const Font&o):GdiObj(o.h) {const_cast<Font&>(o).h=0;}		// Raubkopie
  void operator=(const Font&o) {swap(h,const_cast<Font&>(o).h);}// Tausch statt Zuweisung
  Font(HWND w):GdiObj(GetWindowFont(w)) {}	// avoid destruction by Font::release() before scope leaving!
  Font(HDC dc):GdiObj(GetCurrentObject(dc,OBJ_FONT)) {}
  static HFONT create(const TCHAR*name,int size,DWORD flags=0,int rotation=0) {
    Data d(name,size,flags,rotation);
    return d();	// i.e. CreateFont()
  }
  static HFONT create(const Data&d) {return d();}
  Font(LPCTSTR name,int size,BYTE flags=0,int rotation=0):GdiObj(create(name,size,flags,rotation)) {}
  Font(const Data&d):GdiObj(d()) {}
  Font(const LOGFONT&lf):GdiObj(CreateFontIndirect(&lf)) {}
  void query(LOGFONT&lf) {GetObject(h,sizeof lf,&lf);}
  Data query() {Data&d=*__$ReturnUdt; d(HFONT(h)); return d;}	// get a C++ structure, space is provided by caller
  void operator=(HFONT f) {if (h) DeleteFont(h); h=(HGDIOBJ)f;}
  operator HFONT() const {return (HFONT)h;}	// wenn HFONT gefragt wird
};

// A small metric structure suitable for function return: 32 bits
// All four needed Y values for font processing; extlead only for multi-line text
struct Metric{
  BYTE height,ascent,intlead,extlead;
  Metric(const TEXTMETRIC&tm):
   height(BYTE(tm.tmHeight)),	// Height of "q"
   ascent(BYTE(tm.tmAscent)),	// Height of ""
   intlead(BYTE(tm.tmInternalLeading)),	// Height of "" - "A"
   extlead(BYTE(tm.tmExternalLeading)) {}	// multi-line gap (Durchschuss)
  Metric(HDC dc) {
    TEXTMETRIC tm;
    GetTextMetrics(dc,&tm);
    height=BYTE(tm.tmHeight);
    ascent=BYTE(tm.tmAscent);
    intlead=BYTE(tm.tmInternalLeading);
    extlead=BYTE(tm.tmExternalLeading);
  }
  BYTE descent() const {return height-ascent;}	// Unterlnge
};
